# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

"""Version number."""

NAME = "CBMC starter kit"
NUMBER = "2.3"
VERSION = f"{NAME} {NUMBER}"

REPLACE_TARGET = '_CBMC_STARTER_KIT_VERSION_'

def version(display=False):
    """The version of cbmc viewer."""

    if display:
        print(VERSION)
    return VERSION

def copy_with_version(src, dst):
    with open(src, encoding='utf-8') as srcfile:
        data = srcfile.read()
    with open(dst, 'w', encoding='utf-8') as dstfile:
        dstfile.write(data.replace(REPLACE_TARGET, version()))



"""Ctags support for locating symbol definitions"""

from pathlib import Path
import json
import logging
import subprocess
import sys

################################################################
# This popen method is used to subprocess-out the invocation of ctags.
# This method duplicates code in other modules to make this ctags
# module a stand-alone module that can be copied and reused in other
# projects.

def popen(cmd, cwd=None, stdin=None, encoding=None):
    """Run a command with string stdin on stdin, return stdout and stderr."""

    cmd = [str(word) for word in cmd]
    kwds = {'cwd': cwd,
            'text': True,
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE}
    if sys.version_info >= (3, 6): # encoding is new in Python 3.6
        kwds['encoding'] = encoding or 'utf-8'
    try:
        logging.debug('Popen command: "%s"', ' '.join(cmd))
        logging.debug('Popen stdin: "%s"', stdin)
        with subprocess.Popen(cmd, **kwds) as pipe:
            stdout, stderr = pipe.communicate(input=stdin)
        logging.debug('Popen stdout: "%s"', stdout)
        logging.debug('Popen stderr: "%s"', stderr)
        if pipe.returncode:
            logging.debug('Popen command failed: "%s"', ' '.join(cmd))
            logging.debug('Popen return code: "%s"', pipe.returncode)
            raise UserWarning(f"Failed to run command: {' '.join(cmd)}")
        return stdout, stderr
    except FileNotFoundError as error:
        logging.debug("FileNotFoundError: command '%s': %s", ' '.join(cmd), error)
        raise UserWarning(f"Failed to run command: {' '.join(cmd)}") from error

################################################################

def ctags(root, files):
    """List symbols defined in files under root."""

    root = Path(root)
    files = [str(file_) for file_ in files]
    return (universal_ctags(root, files) or
            exhuberant_ctags(root, files) or
            legacy_ctags(root, files) or
            [])

################################################################

def universal_ctags(root, files):
    """Use universal ctags to list symbols defined in files under root."""

    # See universal ctags man page at https://docs.ctags.io/en/latest/man/ctags.1.html
    cmd = [
        'ctags',
        '-L', '-', # read files from standard input, one file per line
        '-f', '-', # write tags to standard output, one tag per line
        '--output-format=json', # each tag is a one-line json blob
        '--fields=FNnK' # json blob is {"name": symbol, "path": file, "line": line, "kind": kind}
    ]
    try:
        logging.info("Running universal ctags")
        stdout, _ = popen(cmd, cwd=root, stdin='\n'.join(files))
        strings = stdout.splitlines()
    except UserWarning:
        logging.info("Universal ctags failed")
        strings = []

    return [tag for string in strings for tag in universal_tag(root, string)]

def universal_tag(root, string):
    """Extract tag from universal ctag output."""

    try:
        # universal ctag json output is '{"name": symbol, "path": file, "line": line, "kind": kind}'
        blob = json.loads(string)
        return [{'symbol': blob['name'], 'file': root/blob['path'], 'line': int(blob['line']),
                 'kind': blob['kind']}]
    except (json.decoder.JSONDecodeError, # json is unparsable
            KeyError,                     # json key is missing
            ValueError) as error:         # invalid literal for int()
        logging.debug("Bad universal ctag: %s: %s", string, error)
        return []

################################################################

def exhuberant_ctags(root, files):
    """Use exhuberant ctags to list symbols defined in files under root."""

    # See exhuberant ctags man page at https://linux.die.net/man/1/ctags
    cmd = [
        'ctags',
        '-L', '-', # read files from standard input, one file per line
        '-f', '-', # write tags to standard output, one tag per line
        '-n', # use line numbers (not search expressions) to locate symbol in file
        '--fields=K' # include symbol kind among extension fields
    ]
    try:
        logging.info("Running exhuberant ctags")
        stdout, _ = popen(cmd, cwd=root, stdin='\n'.join(files))
        strings = stdout.splitlines()
    except UserWarning:
        logging.info("Exhuberant ctags failed")
        strings = []

    return [tag for string in strings for tag in exhuberant_tag(root, string)]

def exhuberant_tag(root, string):
    """Extract tag from exhuberant ctag output."""

    try:
        # exhuberant ctag output is 'symbol<TAB>path<TAB>line;"<TAB>kind'
        left, right = string.split(';"')[:2]
        symbol, path, line = left.split("\t")[:3]
        kind = right.split("\t")[1]
        return [{'symbol': symbol, 'file': root/path, 'line': int(line), 'kind': kind}]
    except (ValueError, IndexError): # not enough values to unpack, invalid literal for int()
        logging.debug('Bad exhuberant ctag: "%s"', string)
        return []

################################################################

def legacy_ctags(root, files):
    """Use legacy ctags to list symbols defined in files under root."""

    # MacOS ships with a legacy ctags from BSD installed in /usr/bin/ctags.
    # See the MacOS man page for the documentation used to implement this method.
    cmd = ['ctags',
           '-x',  # write human-readable summary to standard output
           *files # legacy ctags cannot read list of files from stdin
    ]
    try:
        logging.info("Running legacy ctags")
        stdout, _ = popen(cmd, cwd=root)
        strings = stdout.splitlines()
    except UserWarning:
        logging.info("Legacy ctags failed")
        strings = []
    return [tag for string in strings for tag in legacy_tag(root, string)]

def legacy_tag(root, string):
    """Extract tag from legacy ctag output."""

    try:
        # legacy ctag -x output is 'symbol line path source_code_fragment'
        symbol, line, path = string.split()[:3]
        return [{'symbol': symbol, 'file': root/path, 'line': int(line), 'kind': None}]
    except ValueError: # not enough values to unpack, invalid literal for int()
        logging.debug('Bad legacy ctag: "%s"', string)
        return []

################################################################



#!/usr/bin/env python3

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

"""Set up the CBMC proof instrastructure."""

from pathlib import Path
import os
import shutil

from cbmc_starter_kit import arguments
from cbmc_starter_kit import repository
from cbmc_starter_kit import update
from cbmc_starter_kit import util

################################################################

def parse_arguments():
    desc = "Set up CBMC proof infrastructure for a repository."
    options = []
    args = arguments.create_parser(options, desc).parse_args()
    arguments.configure_logging(args)
    return args

################################################################

SRCDIR_TEXT = """
# Absolute path to the root of the source tree.
#
SRCDIR ?= {}
"""

LITANI_TEXT = """
# Absolute path to the litani script.
#
LITANI ?= {}
"""

PROJECT_TEXT = """
# Name of this proof project, displayed in proof reports. For example,
# "s2n" or "Amazon FreeRTOS". For projects with multiple proof roots,
# this may be overridden on the command-line to Make, for example
#
#         make PROJECT_NAME="FreeRTOS MQTT" report
#
PROJECT_NAME = "{}"
"""

def srcdir_definition(source_root, proof_root):
    # Let makefile construct absolute path to source root
    srcdir_path = f"$(abspath $(PROOF_ROOT)/{os.path.relpath(source_root, proof_root)})"
    return SRCDIR_TEXT.format(srcdir_path)

def litani_definition(litani, proof_root):
    if litani.is_file():
        # Let makefile construct absolute path to litani
        litani_path = f"$(abspath $(PROOF_ROOT)/{os.path.relpath(litani, proof_root)})"
        return LITANI_TEXT.format(litani_path)
    return LITANI_TEXT.format(litani)

def project_name_definition(project_name):
    return PROJECT_TEXT.format(project_name)

################################################################

def main():
    """Set up the CBMC proof infrastructure."""

    parse_arguments() # only arguments are --verbose and --debug

    # Gather project-specific definitions
    source_root = repository.repository_root()
    if shutil.which("litani"):
        litani = Path("litani")
    else:
        litani = repository.litani_root() / "litani"
    project_name = util.ask_for_project_name()

    # Copy cbmc infrastructure into cbmc directory
    cbmc_root = Path.cwd()
    shutil.copytree(util.package_repository_template_root(), cbmc_root, dirs_exist_ok=True)
    shutil.rmtree(cbmc_root / util.NEGATIVE_TESTS, ignore_errors=True)
    shutil.rmtree(cbmc_root / util.PROOF_DIR / "__pycache__", ignore_errors=True)

    # Overwrite Makefile.common and run-cbmc-proofs.py with versioned copies
    # Quiet warnings about overwriting files
    update.update(cbmc_root, quiet=True)

    # Write project-specific definitions to cbmc/proofs/Makefile-template-defines
    proof_root = cbmc_root / util.PROOF_DIR
    makefile = proof_root/util.TEMPLATE_DEFINES
    with open(makefile, "w", encoding='utf-8') as mkfile:
        print(srcdir_definition(source_root, proof_root), file=mkfile)
        print(litani_definition(litani, proof_root), file=mkfile)
        print(project_name_definition(project_name), file=mkfile)

if __name__ == "__main__":
    main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

"""Methods for common command-line argument parsing."""

import argparse
import logging

from cbmc_starter_kit import version as starter_kit_version

def create_parser(options=None, description=None, epilog=None):
    """Create a parser for command line arguments."""

    options = options or []
    description = description or ""

    flags = [option.get('flag') for option in options]
    if '--verbose' not in flags:
        options.append({'flag': '--verbose', 'action': 'store_true', 'help': 'Verbose output'})
    if '--debug' not in flags:
        options.append({'flag': '--debug', 'action': 'store_true', 'help': 'Debug output'})
    if '--version' not in flags:
        options.append({'flag': '--version',
                        'action': 'version', 'version': starter_kit_version.version(),
                        'help': 'Display version and exit'})

    parser = argparse.ArgumentParser(description=description, epilog=epilog)
    for option in options:
        flag = option.pop('flag')
        parser.add_argument(flag, **option)
    return parser

def configure_logging(args):
    """Configure logging level based on command line arguments."""

    # Logging is configured by first invocation of basicConfig
    fmt = '%(levelname)s: %(message)s'
    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format=fmt)
        return
    if args.verbose:
        logging.basicConfig(level=logging.INFO, format=fmt)
        return
    logging.basicConfig(format=fmt)
